package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.DefaultErrorHandlingPolicies;
import org.eclipse.jdt.internal.compiler.IErrorHandlingPolicy;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.ExpressionContext;
import org.eclipse.jdt.internal.compiler.ast.LambdaExpression;
import org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.ImplicitNullAnnotationVerifier;
import org.eclipse.jdt.internal.compiler.lookup.IntersectionTypeBinding18;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;

public abstract class FunctionalExpression extends Expression {
  protected TypeBinding expectedType;
  
  public MethodBinding descriptor;
  
  public MethodBinding binding;
  
  protected MethodBinding actualMethodBinding;
  
  boolean ignoreFurtherInvestigation;
  
  protected ExpressionContext expressionContext = ExpressionContext.VANILLA_CONTEXT;
  
  public CompilationResult compilationResult;
  
  public BlockScope enclosingScope;
  
  public int bootstrapMethodNumber = -1;
  
  public boolean shouldCaptureInstance = false;
  
  protected static IErrorHandlingPolicy silentErrorHandlingPolicy = DefaultErrorHandlingPolicies.ignoreAllProblems();
  
  private boolean hasReportedSamProblem = false;
  
  public boolean isSerializable;
  
  public FunctionalExpression(CompilationResult compilationResult) {
    this.compilationResult = compilationResult;
  }
  
  public FunctionalExpression() {}
  
  public boolean isBoxingCompatibleWith(TypeBinding targetType, Scope scope) {
    return false;
  }
  
  public void setCompilationResult(CompilationResult compilationResult) {
    this.compilationResult = compilationResult;
  }
  
  public MethodBinding getMethodBinding() {
    return null;
  }
  
  public void setExpectedType(TypeBinding expectedType) {
    this.expectedType = expectedType;
  }
  
  public void setExpressionContext(ExpressionContext context) {
    this.expressionContext = context;
  }
  
  public ExpressionContext getExpressionContext() {
    return this.expressionContext;
  }
  
  public boolean isPolyExpression(MethodBinding candidate) {
    return true;
  }
  
  public boolean isPolyExpression() {
    return true;
  }
  
  public boolean isFunctionalType() {
    return true;
  }
  
  public boolean isPertinentToApplicability(TypeBinding targetType, MethodBinding method) {
    if (targetType instanceof TypeVariableBinding) {
      TypeVariableBinding typeVariable = (TypeVariableBinding)targetType;
      if (method != null) {
        if (typeVariable.declaringElement == method)
          return false; 
        if (method.isConstructor() && typeVariable.declaringElement == method.declaringClass)
          return false; 
      } else if (typeVariable.declaringElement instanceof MethodBinding) {
        return false;
      } 
    } 
    return true;
  }
  
  public TypeBinding invocationTargetType() {
    if (this.expectedType == null)
      return null; 
    MethodBinding sam = this.expectedType.getSingleAbstractMethod((Scope)this.enclosingScope, true);
    if (sam != null && sam.problemId() != 17) {
      if (sam.isConstructor())
        return (TypeBinding)sam.declaringClass; 
      return sam.returnType;
    } 
    return null;
  }
  
  public TypeBinding expectedType() {
    return this.expectedType;
  }
  
  public boolean argumentsTypeElided() {
    return true;
  }
  
  public int recordFunctionalType(Scope scope) {
    while (scope != null) {
      ReferenceContext context;
      CompilationUnitDeclaration unit;
      switch (scope.kind) {
        case 2:
          context = ((MethodScope)scope).referenceContext;
          if (context instanceof LambdaExpression) {
            LambdaExpression expression = (LambdaExpression)context;
            if (expression != expression.original)
              return 0; 
          } 
          break;
        case 4:
          unit = ((CompilationUnitScope)scope).referenceContext;
          return unit.record(this);
      } 
      scope = scope.parent;
    } 
    return 0;
  }
  
  public TypeBinding resolveType(BlockScope blockScope) {
    return resolveType(blockScope, false);
  }
  
  public TypeBinding resolveType(BlockScope blockScope, boolean skipKosherCheck) {
    this.constant = Constant.NotAConstant;
    this.enclosingScope = blockScope;
    MethodBinding sam = (this.expectedType == null) ? null : this.expectedType.getSingleAbstractMethod((Scope)blockScope, argumentsTypeElided());
    if (sam == null) {
      blockScope.problemReporter().targetTypeIsNotAFunctionalInterface(this);
      return null;
    } 
    if (!sam.isValidBinding())
      return reportSamProblem(blockScope, sam); 
    this.descriptor = sam;
    if (skipKosherCheck || kosherDescriptor((Scope)blockScope, sam, true)) {
      if (this.expectedType instanceof IntersectionTypeBinding18) {
        ReferenceBinding[] intersectingTypes = ((IntersectionTypeBinding18)this.expectedType).intersectingTypes;
        for (int t = 0, max = intersectingTypes.length; t < max; t++) {
          if (intersectingTypes[t].findSuperTypeOriginatingFrom(37, false) != null) {
            this.isSerializable = true;
            break;
          } 
        } 
      } else if (this.expectedType.findSuperTypeOriginatingFrom(37, false) != null) {
        this.isSerializable = true;
      } 
      LookupEnvironment environment = blockScope.environment();
      if (environment.globalOptions.isAnnotationBasedNullAnalysisEnabled) {
        if ((sam.tagBits & 0x1000L) == 0L)
          (new ImplicitNullAnnotationVerifier(environment, environment.globalOptions.inheritNullAnnotations))
            .checkImplicitNullAnnotations(sam, null, false, (Scope)blockScope); 
        NullAnnotationMatching.checkForContradictions(sam, this, (Scope)blockScope);
      } 
      return this.resolvedType = this.expectedType;
    } 
    return this.resolvedType = null;
  }
  
  protected TypeBinding reportSamProblem(BlockScope blockScope, MethodBinding sam) {
    if (this.hasReportedSamProblem)
      return null; 
    switch (sam.problemId()) {
      case 17:
        blockScope.problemReporter().targetTypeIsNotAFunctionalInterface(this);
        this.hasReportedSamProblem = true;
        break;
      case 18:
        blockScope.problemReporter().illFormedParameterizationOfFunctionalInterface(this);
        this.hasReportedSamProblem = true;
        break;
    } 
    return null;
  }
  
  public boolean kosherDescriptor(Scope scope, MethodBinding sam, boolean shouldChatter) {
    VisibilityInspector inspector = new VisibilityInspector(this, this, scope, shouldChatter);
    boolean status = true;
    if (!inspector.visible(sam.returnType))
      status = false; 
    if (!inspector.visible(sam.parameters))
      status = false; 
    if (!inspector.visible((TypeBinding[])sam.thrownExceptions))
      status = false; 
    if (!inspector.visible(this.expectedType))
      status = false; 
    return status;
  }
  
  public int nullStatus(FlowInfo flowInfo) {
    return 4;
  }
  
  public int diagnosticsSourceEnd() {
    return this.sourceEnd;
  }
  
  public MethodBinding[] getRequiredBridges() {
    ReferenceBinding functionalType;
    if (this.expectedType instanceof IntersectionTypeBinding18) {
      functionalType = (ReferenceBinding)((IntersectionTypeBinding18)this.expectedType).getSAMType((Scope)this.enclosingScope);
    } else {
      functionalType = (ReferenceBinding)this.expectedType;
    } 
    return (new BridgeCollector(this, functionalType, this.descriptor)).getBridges();
  }
  
  boolean requiresBridges() {
    return (getRequiredBridges() != null);
  }
}
